import itertools
import os
import random
from collections.abc import Iterable
from typing import Any

import cv2
import numpy as np
import pandas as pd
from tqdm import tqdm

MOT_COLUMNS = [
    "frame_id",
    "track_id",
    "xmin",
    "ymin",
    "width",
    "height",
    "conf",
    "x",
    "y",
    "z",
]
COLUMN_SUBSET = ["frame_id", "track_id", "xmin", "ymin", "xmax", "ymax"]

COLORS: list[tuple[int, int, int]] = [
    (random.randint(0, 240), random.randint(0, 240), random.randint(0, 240))
    for _ in range(100)
]


def add_max(df: pd.DataFrame) -> None:
    df["xmax"] = df["xmin"] + df["width"]
    df["ymax"] = df["ymin"] + df["height"]


def load_csv(base_path: str, dataset: str, ignore_conf0: bool = False) -> pd.DataFrame:
    csv_path = os.path.join(base_path, dataset, f"{dataset}.txt")
    img_pattern = os.path.join(base_path, "img1", "{:06}.jpg")

    df = pd.read_csv(csv_path, header=None, names=MOT_COLUMNS)
    add_max(df)
    if ignore_conf0:
        df = df[df["conf"] != 0]

    df["xmin"] = df["xmin"].astype(float)
    df["ymin"] = df["ymin"].astype(float)
    add_max(df)
    df = df[COLUMN_SUBSET].copy()
    df["img_path"] = df["frame_id"].map(lambda i: img_pattern.format(i))

    # no (reliable) score, so just add 1.0
    df["score"] = 1.0
    df.set_index("frame_id", inplace=True)
    return df


def read_img(path: str) -> np.ndarray:
    return cv2.imread(path)[..., ::-1]


def plot_bbox(
    img: np.ndarray,
    xmin: float,
    ymin: float,
    xmax: float,
    ymax: float,
    color: tuple[int, int, int] = (255, 0, 0),
) -> np.ndarray:
    return cv2.rectangle(
        img,
        (int(round(xmin)), int(round(ymin))),
        (int(round(xmax)), int(round(ymax))),
        color,
        2,
    )


def get_annotated_frame(frame_boxes: pd.DataFrame) -> np.ndarray:
    assert frame_boxes["img_path"].nunique() == 1
    img = read_img(frame_boxes["img_path"].iloc[0]).copy()
    for b in frame_boxes.itertuples():
        plot_bbox(
            img,
            b.xmin,
            b.ymin,
            b.xmax,
            b.ymax,
            color=COLORS[int(b.track_id) % len(COLORS)],
        )
    return img


def run_tracker(
    tracker: Any, boxes: pd.DataFrame, return_all: bool = False
) -> pd.DataFrame:
    tracks = []
    frame_paths = boxes["img_path"].to_dict()
    for frame_id in tqdm(range(boxes.index.min(), boxes.index.max() + 1)):
        frame_boxes = boxes.loc[frame_id:frame_id]
        track_boxes = frame_boxes[["xmin", "ymin", "xmax", "ymax", "score"]].values

        frame_tracks = tracker.update(track_boxes, return_all=return_all)
        tracks.append(
            np.concatenate(
                (frame_tracks, frame_id * np.ones((frame_tracks.shape[0], 1))), axis=1
            )
        )

    df = pd.DataFrame(
        np.concatenate(tracks),
        columns=["xmin", "ymin", "xmax", "ymax", "track_id", "frame_id"],
    )
    df["img_path"] = df["frame_id"].map(frame_paths)
    return df.set_index("frame_id")


def write_video(out_path: str, fps: int, frames: Iterable[np.ndarray]) -> None:
    frame_iterator = iter(frames)
    frame = next(frame_iterator)
    h, w, _ = frame.shape
    fourcc = cv2.VideoWriter_fourcc("m", "p", "4", "v")
    writer = cv2.VideoWriter(out_path, fourcc, fps, (w, h))

    for frame in itertools.chain((frame,), frame_iterator):
        writer.write(frame[..., ::-1])

    writer.release()



import numpy as np
import numpy.typing as npt
from typing import Union

class BaseTracker:
    def __new__(self) -> BaseTracker: ...
    def update(
        self, boxes: npt.NDArray[Union[np.float32, np.float64]], return_all: bool
    ) -> npt.NDArray[np.float32]: ...
    def get_current_track_boxes(
        self, return_all: bool = False
    ) -> npt.NDArray[np.float32]: ...
    def clear_trackers(self) -> None: ...
    def remove_tracker(self, track_id: int) -> None: ...

class Sort(BaseTracker):
    max_age: int
    min_hits: int
    iou_threshold: float
    init_tracker_min_score: float
    tracklets: dict[int, KalmanBoxTracker]
    n_steps: int

    def __new__(
        self,
        max_age: int = 1,
        min_hits: int = 3,
        iou_threshold: float = 0.3,
        init_tracker_min_score: float = 0.0,
        measurement_noise: tuple[float, float, float, float] = (1.0, 1.0, 10.0, 0.05),
        process_noise: tuple[float, float, float, float, float, float, float] = (
            1.0,
            1.0,
            1.0,
            0.001,
            0.01,
            0.01,
            0.0001,
        ),
    ) -> Sort:
        """Create a new SORT bbox tracker

        Parameters
        ----------
        max_age
            maximum frames a tracklet is kept alive without matching detections
        min_hits
            minimum number of successive detections before a tracklet is set to alive
        iou_threshold
            minimum IOU to assign detection to tracklet
        init_tracker_min_score
            minimum score to create a new tracklet from unmatched detection box
        measurement_noise
            Diagonal of the measurement noise covariance matrix
            i.e. uncertainties of (x, y, s, r) measurements
            defaults should be reasonable in most cases
         process_noise
            Diagonal of the process noise covariance matrix
            i.e. uncertainties of (x, y, s, r, dx, dy, ds) during each step
            defaults should be reasonable in most cases
        """
        ...
    def update(
        self, boxes: npt.NDArray[Union[np.float32, np.float64]], return_all: bool
    ) -> npt.NDArray[np.float32]:
        """Update the tracker with new boxes and return position of current tracklets

        Parameters
        ----------
        boxes
            array of boxes of shape (n_boxes, 5)
            of the form [[xmin1, ymin1, xmax1, ymax1, score1], [xmin2,...],...]
        return_all
            if true return all living trackers, including inactive (but not dead) ones
            otherwise return only active trackers (those that got at least min_hits
            matching boxes in a row)

        Returns
        -------
            array of tracklet boxes with shape (n_tracks, 5)
            of the form [[xmin1, ymin1, xmax1, ymax1, track_id1], [xmin2,...],...]
        """
        ...
    def get_current_track_boxes(
        self, return_all: bool = False
    ) -> npt.NDArray[np.float32]:
        """Return current track boxes

        Parameters
        ----------
        return_all
            if true return all living trackers, including inactive (but not dead) ones
            otherwise return only active trackers (those that got at least min_hits
            matching boxes in a row)

        Returns
        -------
           array of tracklet boxes with shape (n_tracks, 5)
           of the form [[xmin1, ymin1, xmax1, ymax1, track_id1], [xmin2,...],...]
        """
        ...
    def clear_trackers(self) -> None:
        """Remove all tracklets"""
        ...
    def remove_tracker(self, track_id: int) -> None:
        """Remove tracklet with the given track_id,
        do nothing if it doesn't exist"""
        ...

class ByteTrack(BaseTracker):
    max_age: int
    min_hits: int
    iou_threshold: float
    init_tracker_min_score: float
    tracklets: dict[int, KalmanBoxTracker]
    n_steps: int

    def __new__(
        self,
        max_age: int = 1,
        min_hits: int = 3,
        iou_threshold: float = 0.3,
        init_tracker_min_score: float = 0.8,
        high_score_threshold: float = 0.7,
        low_score_threshold: float = 0.1,
        measurement_noise: tuple[float, float, float, float] = (1.0, 1.0, 10.0, 0.05),
        process_noise: tuple[float, float, float, float, float, float, float] = (
            1.0,
            1.0,
            1.0,
            0.001,
            0.01,
            0.01,
            0.0001,
        ),
    ) -> ByteTrack:
        """Create a new SORT bbox tracker

        Parameters
        ----------
        max_age
            maximum frames a tracklet is kept alive without matching detections
        min_hits
            minimum number of successive detections before a tracklet is set to alive
        iou_threshold
            minimum IOU to assign detection to tracklet
        init_tracker_min_score
            minimum score to create a new tracklet from unmatched detection box
        high_score_threshold
            boxes with higher scores than this will be used in the first round of association
        low_score_threshold
            boxes with score between low_score_threshold and high_score_threshold
            will be used in the second round of association
        measurement_noise
            Diagonal of the measurement noise covariance matrix
            i.e. uncertainties of (x, y, s, r) measurements
            defaults should be reasonable in most cases
        process_noise
            Diagonal of the process noise covariance matrix
            i.e. uncertainties of (x, y, s, r, dx, dy, ds) during each step
            defaults should be reasonable in most cases
        """
        ...
    def update(
        self, boxes: npt.NDArray[Union[np.float32, np.float64]], return_all: bool
    ) -> npt.NDArray[np.float32]:
        """Update the tracker with new boxes and return position of current tracklets

        Parameters
        ----------
        boxes
            array of boxes of shape (n_boxes, 5)
            of the form [[xmin1, ymin1, xmax1, ymax1, score1], [xmin2,...],...]
        return_all
            if true return all living trackers, including inactive (but not dead) ones
            otherwise return only active trackers (those that got at least min_hits
            matching boxes in a row)

        Returns
        -------
            array of tracklet boxes with shape (n_tracks, 5)
            of the form [[xmin1, ymin1, xmax1, ymax1, track_id1], [xmin2,...],...]
        """
        ...
    def get_current_track_boxes(
        self, return_all: bool = False
    ) -> npt.NDArray[np.float32]:
        """Return current track boxes

        Parameters
        ----------
        return_all
            if true return all living trackers, including inactive (but not dead) ones
            otherwise return only active trackers (those that got at least min_hits
            matching boxes in a row)

        Returns
        -------
           array of tracklet boxes with shape (n_tracks, 5)
           of the form [[xmin1, ymin1, xmax1, ymax1, track_id1], [xmin2,...],...]
        """
        ...
    def clear_trackers(self) -> None:
        """Remove all tracklets"""
        ...
    def remove_tracker(self, track_id: int) -> None:
        """Remove tracklet with the given track_id,
        do nothing if it doesn't exist"""
        ...

class Bbox:
    xmin: float
    ymin: float
    xmax: float
    ymax: float

class KalmanBoxTracker:
    id: int
    age: int
    hits: int
    hit_streak: int
    steps_since_hit: int
